#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          02-07-2018
//      Created for Project :   FLAME
"""

import pytest
from json import dumps
from pyramid import testing
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound, HTTPConflict
from clmcservice.models import ServiceFunctionChain
from clmcservice.configapi.views import SFCConfigAPI


class TestSFCConfigAPI(object):
    """
    A pytest-implementation test for the Config API endpoints for service function chains
    """

    @pytest.fixture(autouse=True)
    def app_config(self):
        """
        A fixture to implement setUp/tearDown functionality for all tests by initializing configuration structure for the web service and db connection
        """

        self.registry = testing.setUp()

        yield

        testing.tearDown()
        ServiceFunctionChain.delete_all()  # clear the instances of the model in the test database

    def test_get_all(self):
        """
        Tests the GET all method of the config API for service function chains - returns a list of all service function chains from the database.
        """

        request = testing.DummyRequest()
        response = SFCConfigAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function chains in the database."

        sfc = ServiceFunctionChain(sfc="sfc1", sfc_i="sfc_i1", chain={"nginx": ["minio"]})
        expected_response_data = [sfc.json]
        ServiceFunctionChain.add(sfc)  # adds the new instance of the model to the database

        request = testing.DummyRequest()
        response = SFCConfigAPI(request).get_all()
        assert response == expected_response_data, "Incorrect response data with 1 service function chain."

        sfc = ServiceFunctionChain(sfc="sfc2", sfc_i="sfc_i1", chain={"nginx": ["minio"]})
        expected_response_data.append(sfc.json)
        ServiceFunctionChain.add(sfc)
        sfc = ServiceFunctionChain(sfc="sfc2", sfc_i="sfc_i2", chain={"nginx": ["minio"]})
        expected_response_data.append(sfc.json)
        ServiceFunctionChain.add(sfc)

        request = testing.DummyRequest()
        response = SFCConfigAPI(request).get_all()
        assert response == expected_response_data, "Incorrect response data with more than 1 service function chains."

    def test_get_one(self):
        """
        Tests the GET one method of the config API for service function chains - returns an instance of a service function chain from the database.
        """

        request = testing.DummyRequest()
        response = SFCConfigAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function chains in the database."

        self._validation_of_url_parameters_test("get_one")

        sfc = ServiceFunctionChain(sfc="sfc1", sfc_i="sfc_i1", chain={"nginx": ["minio"]})
        expected_response_data = sfc.json
        ServiceFunctionChain.add(sfc)  # adds the new instance of the model to the database

        request = testing.DummyRequest()
        request.params["sfc"] = "sfc1"
        request.params["sfc_i"] = "sfc_i1"
        response = SFCConfigAPI(request).get_one()
        assert response == expected_response_data, "Invalid data returned in the response of GET instance"

        request = testing.DummyRequest()
        request.params["sfc"] = "sfc2"
        request.params["sfc_i"] = "sfc_i2"
        error_raised = False
        try:
            SFCConfigAPI(request).get_one()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "Not found error must be raised in case of a non existing service function chain"

    def test_post(self):
        """
        Tests the POST method of the config API for service function chains - creates an instance of a service function chain in the database.
        """

        request = testing.DummyRequest()
        response = SFCConfigAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function chains in the database."

        resource = dict(sfc="sfc1", sfc_i="sfc_i1", chain={"nginx": ["minio"]})
        json_data = dumps(resource)
        request = testing.DummyRequest()
        request.body = json_data.encode(request.charset)
        response = SFCConfigAPI(request).post()
        assert response == resource, "POST request must return the created resource"
        assert ServiceFunctionChain.exists("sfc1", "sfc_i1"), "POST request must have created the resource"

        resource["chain"] = {}
        json_data = dumps(resource)
        request = testing.DummyRequest()
        request.body = json_data.encode(request.charset)
        error_raised = False
        try:
            SFCConfigAPI(request).post()
        except HTTPConflict:
            error_raised = True
        assert error_raised, "An error must be raised when trying to create a resource which breaks the unique constraint"

    @pytest.mark.parametrize("body, valid", [
        ('{"sfc": "sfc1", "sfc_i": "sfc_i1", "chain":{"nginx":["minio"]}}', True),
        ('{"sfc": "sfc2", "sfc_i": "sfc_i2", "chain":{}}', True),
        ('{"sfc": "sfc1", "sfc_i": "sfc_i1", "chain":[]}', False),
        ('{}', False),
        ('{"sfc": "sfc3", "sfc_i": "sfc_i3"', False),
        ('{"sf": "sfc2", "sf_i": "sfc_i2", "chain":{}', False),
        ('{invalid json}', False),
    ])
    def test_post_body_validation(self, body, valid):
        """
        Tests the POST request validation of the body content.

        :param body: The request body to be validated
        :param valid: True if body is valid, False otherwise
        """

        request = testing.DummyRequest()
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            SFCConfigAPI(request).post()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised == (not valid), "An error must be raised in case of an invalid request body"

    def test_put(self):
        """
        Tests the PUT method of the Config API for service function chains - overwrites an instance of a service function chain from the database.
        """

        request = testing.DummyRequest()
        response = SFCConfigAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function chains in the database."

        self._validation_of_url_parameters_test("put")

        resource = dict(sfc="sfc1", sfc_i="sfc_i1", chain={"nginx": ["minio"]})
        body = dumps(resource)
        request = testing.DummyRequest()
        request.params["sfc"] = "sfc1"
        request.params["sfc_i"] = "sfc_i1"
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            SFCConfigAPI(request).put()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "Not found error must be raised in case of a non existing service function chain"

        sfc = ServiceFunctionChain(sfc="sfc1", sfc_i="sfc_i1", chain={"nginx": ["minio"]})
        ServiceFunctionChain.add(sfc)  # adds the new instance of the model to the database

        resource = dict(sfc="sfc1", sfc_i="sfc_i1", chain={})
        body = dumps(resource)
        request = testing.DummyRequest()
        request.params["sfc"] = "sfc1"
        request.params["sfc_i"] = "sfc_i1"
        request.body = body.encode(request.charset)
        response = SFCConfigAPI(request).put()
        assert response == resource, "PUT request must return the updated resource"
        assert ServiceFunctionChain.get("sfc1", "sfc_i1").json["chain"] == {}

        resource = dict(sfc="sfc2", sfc_i="sfc_i2", chain={"nginx": ["minio"]})
        body = dumps(resource)
        request = testing.DummyRequest()
        request.params["sfc"] = "sfc1"
        request.params["sfc_i"] = "sfc_i1"
        request.body = body.encode(request.charset)
        response = SFCConfigAPI(request).put()
        assert response == resource, "PUT request must return the updated resource"
        assert not ServiceFunctionChain.exists("sfc1", "sfc_i1"), "Resource has not been updated"
        assert ServiceFunctionChain.exists("sfc2", "sfc_i2"), "Resource has not been updated"

        sfc = ServiceFunctionChain(sfc="sfc1", sfc_i="sfc_i1", chain={"nginx": ["minio"]})
        ServiceFunctionChain.add(sfc)  # adds the new instance of the model to the database

        resource = dict(sfc="sfc2", sfc_i="sfc_i2", chain={"nginx": ["minio"]})
        body = dumps(resource)
        request = testing.DummyRequest()
        request.params["sfc"] = "sfc1"
        request.params["sfc_i"] = "sfc_i1"
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            SFCConfigAPI(request).put()
        except HTTPConflict:
            error_raised = True
        assert error_raised, "PUT request breaks unique constraint"

    @pytest.mark.parametrize("body, valid", [
        ('{"sfc": "sfc1", "sfc_i": "sfc_i1", "chain":{"nginx":["minio"]}}', True),
        ('{"sfc": "sfc2", "sfc_i": "sfc_i2", "chain":{}}', True),
        ('{"sfc": "sfc1", "sfc_i": "sfc_i1", "chain":[]}', False),
        ('{}', False),
        ('{"sfc": "sfc3", "sfc_i": "sfc_i3"', False),
        ('{"sf": "sfc2", "sf_i": "sfc_i2", "chain":{}', False),
        ('{invalid json}', False),
    ])
    def test_put_body_validation(self, body, valid):
        """
        Tests the PUT request validation of the body content.

        :param body: The request body to be validated
        :param valid: True if body is valid, False otherwise
        """

        sfc = ServiceFunctionChain(sfc="sfc1", sfc_i="sfc_i1", chain={"nginx": ["minio"]})
        ServiceFunctionChain.add(sfc)  # adds the new instance of the model to the database

        request = testing.DummyRequest()
        request.params["sfc"] = "sfc1"
        request.params["sfc_i"] = "sfc_i1"
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            SFCConfigAPI(request).put()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised == (not valid), "An error must be raised in case of an invalid request body"

    def test_delete(self):
        """
        Tests the DELETE method of the config API for service function chains - deletes an instance of a service function chain from the database.
        """

        request = testing.DummyRequest()
        response = SFCConfigAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function chains in the database."

        self._validation_of_url_parameters_test("delete")

        sfc = ServiceFunctionChain(sfc="sfc1", sfc_i="sfc_i1", chain={"nginx": ["minio"]})
        to_delete = sfc.json
        ServiceFunctionChain.add(sfc)  # adds the new instance of the model to the database

        assert ServiceFunctionChain.exists("sfc1", "sfc_i1")

        request = testing.DummyRequest()
        request.params["sfc"] = "sfc1"
        request.params["sfc_i"] = "sfc_i1"
        response = SFCConfigAPI(request).delete()
        assert response == to_delete, "DELETE must return the deleted object if successful"

        assert not ServiceFunctionChain.exists("sfc1", "sfc_i1"), "Resource must be deleted after the delete API method has been called."

        request = testing.DummyRequest()
        request.params["sfc"] = "sfc1"
        request.params["sfc_i"] = "sfc_i1"
        error_raised = False
        try:
            SFCConfigAPI(request).delete()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "Not found error must be raised in case of a non existing service function chain"

    @staticmethod
    def _validation_of_url_parameters_test(method):
        """
        Validates the way a config API method handles url query parameters for service function chains

        :param method: the method to test
        """

        request = testing.DummyRequest()
        error_raised = False
        try:
            getattr(SFCConfigAPI(request), method).__call__()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "Error must be raised in case of no URL parameters"

        request = testing.DummyRequest()
        request.params["sfc"] = "sfc1"
        try:
            getattr(SFCConfigAPI(request), method).__call__()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "Error must be raised in case of insufficient number of arguments"

        request = testing.DummyRequest()
        request.params["sf"] = "sfc1"  # argument should be sfc
        request.params["sfc_i"] = "sfc_i1"
        try:
            getattr(SFCConfigAPI(request), method).__call__()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "Error must be raised in case of invalid naming of arguments"
